/*
 * libmpeg3.c -- libmpeg3 player plugin, which exploits libmpeg3.
 * (C)Copyright 2000, 2001, 2002 by Hiroshi Takekawa
 * This file is part of Enfle.
 *
 * Last Modified: Sun Dec 31 01:28:12 2006.
 * $Id: libmpeg3.c,v 1.50 2007/04/27 05:55:27 sian Exp $
 *
 * NOTES: 
 *  This plugin is not fully enfle plugin compatible, because stream
 *  cannot be used for input. Stream's path is used as input filename.
 *
 *  Requires libmpeg3 1.2.1 (or later).
 *
 * Enfle is free software; you can redistribute it and/or modify it
 * under the terms of the GNU General Public License version 2 as
 * published by the Free Software Foundation.
 *
 * Enfle is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA 02111-1307, USA
 */

#include <stdio.h>
#include <stdlib.h>
#include <mpeg3/libmpeg3.h>

#define REQUIRE_STRING_H
#define REQUIRE_FATAL
#include "compat.h"
#include "common.h"

#ifndef USE_PTHREAD
#  error pthread is mandatory for libmpeg3 plugin
#endif

#include "enfle/player-plugin.h"

typedef struct _libmpeg3_info {
  mpeg3_t *file;
  Image *p;
  Config *c;
  AudioDevice *ad;
  unsigned char **lines;
  char *y, *u, *v;
  int eof;
  int use_xv;
  int rendering_type;
  pthread_mutex_t update_mutex;
  pthread_cond_t update_cond;
  int nvstreams;
  int nvstream;
  pthread_t video_thread;
  int nastreams;
  int nastream;
  pthread_t audio_thread;
} LibMPEG3_info;

static const unsigned int types =
  (IMAGE_I420 |
   IMAGE_RGBA32 | IMAGE_BGRA32 |
   IMAGE_RGB24 | IMAGE_BGR24 |
   IMAGE_RGB565 | IMAGE_BGR565);

DECLARE_PLAYER_PLUGIN_METHODS;

static PlayerStatus play(Movie *);
static PlayerStatus pause_movie(Movie *);
static PlayerStatus stop_movie(Movie *);

static PlayerPlugin plugin = {
  .type = ENFLE_PLUGIN_PLAYER,
  .name = "LibMPEG3",
  .description = "LibMPEG3 Player plugin version 0.5",
  .author = "Hiroshi Takekawa",

  .identify = identify,
  .load = load
};

ENFLE_PLUGIN_ENTRY(player_libmpeg3)
{
  PlayerPlugin *pp;

  if ((pp = (PlayerPlugin *)calloc(1, sizeof(PlayerPlugin))) == NULL)
    return NULL;
  memcpy(pp, &plugin, sizeof(PlayerPlugin));

  return (void *)pp;
}

ENFLE_PLUGIN_EXIT(player_libmpeg3, p)
{
  free(p);
}

/* for internal use */

static PlayerStatus
load_movie(VideoWindow *vw, Movie *m, Stream *st, Config *c)
{
  LibMPEG3_info *info;
  Image *p;
  unsigned int i;
  int direct_renderable;

  if ((info = calloc(1, sizeof(LibMPEG3_info))) == NULL) {
    show_message("LibMPEG3: play_movie: No enough memory.\n");
    return PLAY_ERROR;
  }

  if ((info->file = mpeg3_open(st->path)) == NULL) {
    free(info);
    return PLAY_ERROR;
  }
  info->c = c;

  pthread_mutex_init(&info->update_mutex, NULL);
  pthread_cond_init(&info->update_cond, NULL);

  /* XXX: can set how many CPUs you want to use. */
  /* mpeg3_set_cpus(info->file, 2); */
  mpeg3_set_mmx(info->file, 1);

  m->has_audio = 0;
  if (mpeg3_has_audio(info->file)) {
    if (m->ap == NULL)
      show_message("Audio not played.\n");
    else {
      info->nastreams = mpeg3_total_astreams(info->file);
      /* XXX: stream should be selectable */
      info->nastream = 0;

      m->sampleformat = _AUDIO_FORMAT_S16_LE;
      m->channels = mpeg3_audio_channels(info->file, info->nastream);
      m->samplerate = mpeg3_sample_rate(info->file, info->nastream);
      m->num_of_samples = mpeg3_audio_samples(info->file, info->nastream);

      show_message("audio(%d streams): format(%d): %d ch %d Hz %d samples\n", info->nastreams, m->sampleformat, m->channels, m->samplerate, m->num_of_samples);
      if (m->ap->bytes_written == NULL)
	show_message("audio sync may be incorrect.\n");
      m->has_audio = 1;
    }
  } else {
    debug_message("No audio streams.\n");
  }

  p = info->p = image_create();
  p->direct_renderable = direct_renderable;

  info->use_xv = 0;
  m->has_video = 1;
  if (!mpeg3_has_video(info->file)) {
    m->has_video = 0;
    m->width = 120;
    m->height = 80;
    rational_set_0(m->framerate);
    m->num_of_frames = 1;
    m->rendering_width = m->width;
    m->rendering_height = m->height;
    show_message("This stream has no video stream.\n");
  } else {
    if ((info->nvstreams = mpeg3_total_vstreams(info->file)) > 1) {
      show_message("There are %d video streams in this whole stream.\n", info->nvstreams);
      show_message("Only the first video stream will be played(so far). Sorry.\n");
    }

    /* XXX: stream should be selectable */
    info->nvstream = 0;

    m->width = mpeg3_video_width(info->file, info->nvstream);
    m->height = mpeg3_video_height(info->file, info->nvstream);
    m->framerate.num = mpeg3_frame_rate(info->file, info->nvstream) * 1000;
    m->framerate.den = 1000;
    m->num_of_frames = mpeg3_video_frames(info->file, info->nvstream);

    m->requested_type = video_window_request_type(vw, m->width, m->height, types, &direct_renderable);
    if (!direct_renderable) {
      show_message_fnc("Cannot render directly...\n");
      return PLAY_ERROR;
    }
    debug_message("LibMPEG3: requested type: %s direct\n", image_type_to_string(m->requested_type));

    if (m->requested_type == _I420 || m->requested_type == _YV12)
      info->use_xv = 1;

    {
      int dw, dh;

      video_window_calc_magnified_size(vw, info->use_xv, m->width, m->height, &dw, &dh);

      if (info->use_xv) {
	m->rendering_width  = m->width;
	m->rendering_height = m->height;
      } else {
	m->rendering_width  = dw;
	m->rendering_height = dh;
      }
    }

    show_message("video(%d streams): s (%d,%d) r (%d,%d) %f fps %d frames\n", info->nvstreams,
		 m->width, m->height, m->rendering_width, m->rendering_height,
		 rational_to_double(m->framerate), m->num_of_frames);

    image_width(p) = m->rendering_width;
    image_height(p) = m->rendering_height;
    p->type = m->requested_type;
    if ((image_rendered_image(p) = memory_create()) == NULL)
      goto error;
    memory_request_type(image_rendered_image(p), video_window_preferred_memory_type(vw));

    if (info->use_xv) {
      p->bits_per_pixel = 12;
      image_bpl(p) = image_width(p) * 1.5;

      if (memory_alloc(image_rendered_image(p), image_bpl(p) * image_height(p)) == NULL)
	goto error;
      if ((info->y = malloc(image_bpl(p) * image_height(p))) == NULL)
	goto error;
      if (p->type == _I420) {
	info->u = info->y + image_width(p) * image_height(p);
	info->v = info->u + ((image_width(p) * image_height(p)) >> 2);
      } else if (p->type == _YV12) {
	info->v = info->y + image_width(p) * image_height(p);
	info->u = info->v + ((image_width(p) * image_height(p)) >> 2);
      } else {
	fatal("Unsupport image type %s\n", image_type_to_string(p->type));
      }
    } else {
      switch (vw->bits_per_pixel) {
      case 32:
	switch (p->type) {
	case _RGBA32:
	  info->rendering_type = MPEG3_RGBA8888;
	  break;
	case _BGRA32:
	  info->rendering_type = MPEG3_BGRA8888;
	  break;
	default:
	  show_message_fnc("requested type is %s.\n", image_type_to_string(p->type));
	  return PLAY_ERROR;
	}
	p->depth = 24;
	p->bits_per_pixel = 32;
	image_bpl(p) = image_width(p) * 4;
	break;
      case 24:
	switch (p->type) {
	case _RGB24:
	  info->rendering_type = MPEG3_RGB888;
	  break;
	case _BGR24:
	  info->rendering_type = MPEG3_BGR888;
	  break;
	default:
	  show_message_fnc("requested type is %s.\n", image_type_to_string(p->type));
	  return PLAY_ERROR;
	}
	p->depth = 24;
	p->bits_per_pixel = 24;
	image_bpl(p) = image_width(p) * 3;
	break;
      case 16:
	switch (p->type) {
	case _RGB565:
	case _BGR565:
	  info->rendering_type = MPEG3_RGB565;
	  break;
	default:
	  show_message_fnc("requested type is %s.\n", image_type_to_string(p->type));
	  return PLAY_ERROR;
	}
	p->depth = 16;
	p->bits_per_pixel = 16;
	image_bpl(p) = image_width(p) * 2;
	break;
      default:
	show_message("Cannot render bpp %d\n", vw->bits_per_pixel);
	return PLAY_ERROR;
      }

      if ((info->lines = calloc(image_height(p), sizeof(unsigned char *))) == NULL)
	goto error;
      /* extra 4 bytes are needed for MMX routine */
      if (memory_alloc(image_rendered_image(p), image_bpl(p) * image_height(p) + 4) == NULL)
	goto error;
      if ((info->lines[0] = malloc(image_bpl(p) * image_height(p))) == NULL)
	goto error;
      for (i = 1; i < image_height(p); i++)
	info->lines[i] = info->lines[0] + i * image_bpl(p);
    }
  }

  m->movie_private = (void *)info;
  m->st = st;
  m->status = _STOP;

  m->initialize_screen(vw, m, m->rendering_width, m->rendering_height);

  return  play(m);

 error:
  free(info);
  return PLAY_ERROR;
}

static void *play_video(void *);
static void *play_audio(void *);

static PlayerStatus
play(Movie *m)
{
  LibMPEG3_info *info = (LibMPEG3_info *)m->movie_private;

  debug_message_fn("()\n");

  switch (m->status) {
  case _PLAY:
  case _RESIZING:
    return PLAY_OK;
  case _PAUSE:
    return pause_movie(m);
  case _STOP:
    m->status = _PLAY;
    break;
  case _UNLOADED:
    return PLAY_ERROR;
  default:
    warning("Unknown status %d\n", m->status);
    return PLAY_ERROR;
  }

  mpeg3_set_frame(info->file, 0, info->nastream);
  mpeg3_set_sample(info->file, 0, info->nvstream);
  m->current_frame = 0;
  m->current_sample = 0;
  timer_start(m->timer);
  info->eof = 0;

  if (m->has_video)
    pthread_create(&info->video_thread, NULL, play_video, m);
  if (m->has_audio)
    pthread_create(&info->audio_thread, NULL, play_audio, m);

  return PLAY_OK;
}

static void *
play_video(void *arg)
{
  Movie *m = arg;
  LibMPEG3_info *info = (LibMPEG3_info *)m->movie_private;
  int decode_error;

  debug_message_fn("()\n");

  if (info->use_xv) {
    while (m->status == _PLAY) {
      if (m->current_frame >= m->num_of_frames) {
	info->eof = 1;
	break;
      }

      pthread_mutex_lock(&info->update_mutex);

      decode_error =
	(mpeg3_read_yuvframe(info->file, info->y, info->u, info->v,
			  0, 0,
			  m->width, m->height,
			  info->nvstream) == -1) ? 0 : 1;
      memcpy(memory_ptr(image_rendered_image(info->p)), info->y, image_bpl(info->p) * image_height(info->p));
      pthread_cond_wait(&info->update_cond, &info->update_mutex);
      pthread_mutex_unlock(&info->update_mutex);
    }
  } else {
    while (m->status == _PLAY) {
      if (m->current_frame >= m->num_of_frames) {
	info->eof = 1;
	break;
      }

      pthread_mutex_lock(&info->update_mutex);

      decode_error =
	(mpeg3_read_frame(info->file, info->lines,
			  0, 0,
			  m->width, m->height,
			  m->rendering_width, m->rendering_height,
			  info->rendering_type, info->nvstream) == -1) ? 0 : 1;
      memcpy(memory_ptr(image_rendered_image(info->p)), info->lines[0], image_bpl(info->p) * image_height(info->p));
      pthread_cond_wait(&info->update_cond, &info->update_mutex);
      pthread_mutex_unlock(&info->update_mutex);
    }
  }

  debug_message_fn(" exiting.\n");
  return (void *)PLAY_OK;
}

#define AUDIO_READ_SIZE 2048
#define AUDIO_WRITE_SIZE 4096

static void *
play_audio(void *arg)
{
  Movie *m = arg;
  LibMPEG3_info *info = (LibMPEG3_info *)m->movie_private;
  AudioDevice *ad;
  short input_buffer[AUDIO_WRITE_SIZE];
  short output_buffer[AUDIO_WRITE_SIZE];
  int samples_to_read;

  debug_message_fn("()\n");

  if ((ad = m->ap->open_device(NULL, info->c)) == NULL) {
    show_message("Cannot open device.\n");
    m->has_audio = 0;
    return (void *)PLAY_ERROR;
  }
  info->ad = ad;

  m->sampleformat_actual = m->sampleformat;
  m->channels_actual = m->channels;
  m->samplerate_actual = m->samplerate;
  if (!m->ap->set_params(ad, &m->sampleformat_actual, &m->channels_actual, &m->samplerate_actual))
    show_message("Some params are set wrong.\n");

  show_message("audio device configured: %d ch %d Hz\n", m->channels, m->samplerate);

  samples_to_read = AUDIO_READ_SIZE;

  while (m->status == _PLAY) {
    if (m->current_sample >= m->num_of_samples)
      break;
    if (m->num_of_samples < m->current_sample + samples_to_read)
      samples_to_read = m->num_of_samples - m->current_sample;
    mpeg3_read_audio(info->file, NULL, input_buffer, 0, samples_to_read, info->nastream);
    if (m->channels == 1) {
      m->ap->write_device(ad, (unsigned char *)input_buffer, samples_to_read * sizeof(short));
    } else {
      int i;

      mpeg3_reread_audio(info->file, NULL, input_buffer + AUDIO_READ_SIZE, 1, samples_to_read, info->nastream);
      for (i = 0; i < samples_to_read; i++) {
	output_buffer[i * 2    ] = input_buffer[i];
	output_buffer[i * 2 + 1] = input_buffer[i + AUDIO_READ_SIZE];
      }
      m->ap->write_device(ad, (unsigned char *)output_buffer, (samples_to_read << 1) * sizeof(short));
    }
    if (m->current_sample == 0) {
      timer_stop(m->timer);
      timer_start(m->timer);
    }
    m->current_sample += samples_to_read;
  }

  m->ap->sync_device(ad);
  m->ap->close_device(ad);
  info->ad = NULL;

  debug_message_fn(" exiting.\n");

  return (void *)PLAY_OK;
}

static int
get_audio_time(Movie *m, AudioDevice *ad)
{
  if (m->has_audio && ad && m->ap->bytes_written)
    return (int)((double)m->ap->bytes_written(ad) / m->samplerate * 500.0 / m->channels);
  return (int)((double)m->current_sample * 1000.0 / m->samplerate);
}

static PlayerStatus
play_main(Movie *m, VideoWindow *vw)
{
  LibMPEG3_info *info = (LibMPEG3_info *)m->movie_private;
  Image *p = info->p;
  int i = 0;
  int video_time, audio_time;
  float rate;

  switch (m->status) {
  case _PLAY:
    break;
  case _RESIZING:
    {
      int dw, dh;

      video_window_resize(vw, m->rendering_width, m->rendering_height);
      video_window_calc_magnified_size(vw, info->use_xv, m->width, m->height, &dw, &dh);

      if (info->use_xv) {
	m->rendering_width  = m->width;
	m->rendering_height = m->height;
      } else {
	m->rendering_width  = dw;
	m->rendering_height = dh;
      }
    }
    m->status = _PLAY;
    break;
  case _PAUSE:
  case _STOP:
    return PLAY_OK;
  case _UNLOADED:
    return PLAY_ERROR;
  default:
    warning("Unknown status %d\n", m->status);
    return PLAY_ERROR;
  }

  if (info->eof) {
    stop_movie(m);
    return PLAY_OK;
  }

  if (!m->has_video)
    return PLAY_OK;

  pthread_mutex_lock(&info->update_mutex);

  m->current_frame = mpeg3_get_frame(info->file, info->nvstream);
  rate = rational_to_double(m->framerate);
  video_time = m->current_frame * 1000 / rate;
  if (m->has_audio) {
    audio_time = get_audio_time(m, info->ad);
    //debug_message("v: %d a=%d (%d frame)\n", (int)timer_get_milli(m->timer), audio_time, m->current_frame);

    /* if too fast to display, wait before render */
    while (video_time > audio_time)
      audio_time = get_audio_time(m, info->ad);

    /* skip if delayed */
    i = (audio_time * rate / 1000) - m->current_frame - 1;
    if (i > 0) {
      debug_message("dropped %d frames\n", i);
      mpeg3_drop_frames(info->file, i, info->nvstream);
    }
  } else {
    /* if too fast to display, wait before render */
    while (video_time > timer_get_milli(m->timer)) ;

    /* skip if delayed */
    i = (timer_get_milli(m->timer) * rate / 1000) - m->current_frame - 1;
    if (i > 0) {
      debug_message("dropped %d frames\n", i);
      mpeg3_drop_frames(info->file, i, info->nvstream);
    }
  }

  /* tell video thread to continue decoding */
  pthread_cond_signal(&info->update_cond);
  pthread_mutex_unlock(&info->update_mutex);

  m->render_frame(vw, m, p);

  return PLAY_OK;
}

static PlayerStatus
pause_movie(Movie *m)
{
  switch (m->status) {
  case _PLAY:
  case _RESIZING:
    m->status = _PAUSE;
    timer_pause(m->timer);
    return PLAY_OK;
  case _PAUSE:
    m->status = _PLAY;
    timer_restart(m->timer);
    return PLAY_OK;
  case _STOP:
    return PLAY_OK;
  case _UNLOADED:
    return PLAY_ERROR;
  default:
    warning("Unknown status %d\n", m->status);
    return PLAY_ERROR;
  }

  return PLAY_ERROR;
}

static PlayerStatus
stop_movie(Movie *m)
{
  LibMPEG3_info *info = (LibMPEG3_info *)m->movie_private;
  void *v, *a;

  debug_message_fn("()\n");

  switch (m->status) {
  case _PLAY:
  case _RESIZING:
    m->status = _STOP;
    timer_stop(m->timer);
    break;
  case _PAUSE:
  case _STOP:
    return PLAY_OK;
  case _UNLOADED:
    return PLAY_ERROR;
  default:
    warning("Unknown status %d\n", m->status);
    return PLAY_ERROR;
  }

  pthread_mutex_lock(&info->update_mutex);
  pthread_cond_signal(&info->update_cond);
  pthread_mutex_unlock(&info->update_mutex);
  if (info->video_thread) {
    pthread_join(info->video_thread, &v);
    info->video_thread = 0;
  }

  if (info->audio_thread) {
    pthread_join(info->audio_thread, &a);
    info->audio_thread = 0;
  }

  return PLAY_OK;
}

static void
unload_movie(Movie *m)
{
  LibMPEG3_info *info = (LibMPEG3_info *)m->movie_private;

  debug_message_fn("()\n");

  stop_movie(m);

  if (info) {
    if (info->lines) {
      if (info->lines[0])
	free(info->lines[0]);
      free(info->lines);
    }
    if (info->y)
      free(info->y);
    if (info->p)
      image_destroy(info->p);

    pthread_mutex_destroy(&info->update_mutex);
    pthread_cond_destroy(&info->update_cond);

    debug_message_fnc("closing libmpeg3 file\n");
    /* close */
    mpeg3_close(info->file);

    debug_message_fnc("freeing info\n");
    free(info);
    m->movie_private = NULL;
    debug_message_fnc("all Ok\n");
  }
}

/* methods */

DEFINE_PLAYER_PLUGIN_IDENTIFY(m, st, c, priv)
{
  if (st->path) {
    if (strlen(st->path) >= 4 && !strcasecmp(st->path + strlen(st->path) - 4, ".mp3"))
      return PLAY_NOT;
    return mpeg3_check_sig(st->path) ? PLAY_OK : PLAY_NOT;
  }

  return PLAY_NOT;
}

DEFINE_PLAYER_PLUGIN_LOAD(vw, m, st, c, priv)
{
  debug_message("libmpeg3 player: load() called\n");

#ifdef IDENTIFY_BEFORE_PLAY
  {
    PlayerStatus status;

    if ((status = identify(m, st, c, priv)) != PLAY_OK)
      return status;
    stream_rewind(st);
  }
#endif

  m->play = play;
  m->play_main = play_main;
  m->pause_movie = pause_movie;
  m->stop = stop_movie;
  m->unload_movie = unload_movie;

  return load_movie(vw, m, st, c);
}
